@@ -1,4 +1,5 @@
 """Liquid template definition."""
+# pylint: disable=too-many-lines
 from __future__ import annotations
 
 import re
@@ -10,24 +11,27 @@ from collections import defaultdict
 from io import StringIO
 from pathlib import Path
 
+from typing import Any
 from typing import Awaitable
 from typing import DefaultDict
 from typing import Dict
-from typing import List
-from typing import Any
 from typing import Iterator
+from typing import List
 from typing import Mapping
 from typing import Optional
 from typing import TextIO
 from typing import Tuple
-from typing import Union
 from typing import TYPE_CHECKING
+from typing import Union
 
+from liquid.ast import BlockNode
 from liquid.ast import ChildNode
 from liquid.ast import Node
 
 from liquid.context import Context
 from liquid.context import ReadOnlyChainMap
+from liquid.context import FutureContext
+from liquid.context import FutureVariableCaptureContext
 from liquid.context import VariableCaptureContext
 
 from liquid.exceptions import TemplateTraversalError, LiquidInterrupt
@@ -35,13 +39,14 @@ from liquid.exceptions import LiquidSyntaxError
 from liquid.exceptions import Error
 from liquid.exceptions import TemplateNotFound
 
-from liquid.expression import Expression, StringLiteral
+from liquid.expression import Expression, StringLiteral, FilteredExpression
 from liquid.expression import Identifier
 from liquid.expression import IdentifierPathElement
 from liquid.expression import IdentifierTuple
 from liquid.expression import Literal
 
 from liquid.output import LimitedStringIO
+from liquid.token import TOKEN_TAG
 
 
 if TYPE_CHECKING:  # pragma: no cover
@@ -266,8 +271,6 @@ class BoundTemplate:
     ) -> TemplateAnalysis:
         """Statically analyze this template and any included/rendered templates.
 
-        Currently we only analyze references to template variables.
-
         :param follow_partials: If ``True``, we will try to load partial templates and
             analyze those templates too. Default's to ``True``.
         :type follow_partials: bool
@@ -297,6 +300,8 @@ class BoundTemplate:
             },
             failed_visits=dict(refs.failed_visits),
             unloadable_partials=dict(refs.unloadable_partials),
+            filters=dict(refs.filters),
+            tags=dict(refs.tags),
         )
 
     async def analyze_async(
@@ -319,6 +324,8 @@ class BoundTemplate:
             },
             failed_visits=dict(refs.failed_visits),
             unloadable_partials=dict(refs.unloadable_partials),
+            filters=dict(refs.filters),
+            tags=dict(refs.tags),
         )
 
     def analyze_with_context(
@@ -352,6 +359,7 @@ class BoundTemplate:
             all_variables=dict(Counter(context.all_references)),
             local_variables=dict(Counter(context.local_references)),
             undefined_variables=dict(Counter(context.undefined_references)),
+            filters=dict(Counter(context.filters)),
         )
 
     async def analyze_with_context_async(
@@ -367,6 +375,7 @@ class BoundTemplate:
             all_variables=dict(Counter(context.all_references)),
             local_variables=dict(Counter(context.local_references)),
             undefined_variables=dict(Counter(context.undefined_references)),
+            filters=dict(Counter(context.filters)),
         )
 
 
@@ -389,6 +398,16 @@ class AwareBoundTemplate(BoundTemplate):
         }
 
 
+class FutureBoundTemplate(BoundTemplate):
+    """A ``BoundTemplate`` subclass configured to use the ``FutureContext`` render context.
+
+    See :class:`liquid.future.Environment`.
+    """
+
+    context_class = FutureContext
+    capture_context_class = FutureVariableCaptureContext
+
+
 class TemplateDrop(Mapping[str, Optional[str]]):
     """Template meta data mapping."""
 
@@ -465,18 +484,27 @@ class ReferencedVariable(str):
         return self._parts
 
 
-Refs = Dict[ReferencedVariable, List[Tuple[str, int]]]
-ReferenceMap = DefaultDict[Identifier, List[Tuple[str, int]]]
-NameMap = DefaultDict[str, List[Tuple[str, int]]]
+# (template_name, line_number).
+Location = Tuple[str, int]
+
+# A mapping of template variables to their (template_name, line_number) locations.
+Refs = Dict[ReferencedVariable, List[Location]]
+
+# A mapping of Identifier expressions to their (template_name, line_number) locations.
+IdentifierMap = DefaultDict[Identifier, List[Location]]
+
+# A mapping of template, tag or filter names to their (template_name, line_number)
+# locations.
+NameRefs = Dict[str, List[Location]]
 
 
 # pylint: disable=too-few-public-methods
 class ContextualTemplateAnalysis:
-    """The result of analyzing a template's variables using
+    """The result of analyzing a template's filters and variables using
     :meth:`BoundTemplate.analyze_with_context`.
 
-    Each of the following properties is a dictionary mapping variable names to the
-    number of times that variable was referenced.
+    Each of the following properties is a dictionary mapping variable or filter names
+    to the number of times that variable was referenced.
 
     :ivar all_variables: All variables references along a path through the template's
         syntax tree.
@@ -486,9 +514,10 @@ class ContextualTemplateAnalysis:
     :ivar undefined_variables: The names of variables that could not be resolved. If a
         name is referenced before it is assigned, it will appear in ``undefined`` and
         ``assigns``.
+    :ivar filters: Names of filters found during contextual analysis.
     """
 
-    __slots__ = ("all_variables", "local_variables", "undefined_variables")
+    __slots__ = ("all_variables", "local_variables", "undefined_variables", "filters")
 
     def __init__(
         self,
@@ -496,22 +525,24 @@ class ContextualTemplateAnalysis:
         all_variables: Dict[str, int],
         local_variables: Dict[str, int],
         undefined_variables: Dict[str, int],
+        filters: Dict[str, int],
     ) -> None:
         self.all_variables = all_variables
         self.local_variables = local_variables
         self.undefined_variables = undefined_variables
+        self.filters = filters
 
 
 # pylint: disable=too-few-public-methods
 class TemplateAnalysis:
-    """The result of analyzing a template's variables using
+    """The result of analyzing a template's tags, filters and variables using
     :meth:`BoundTemplate.analyze`.
 
-    Each of the following properties is a dictionary mapping variable names to a list of
-    two-tuples. Each tuple holds the location of a reference to the name as (<template
-    name>, <line number>). If a name is referenced multiple times, it will appear
-    multiple times in the list. If a name is referenced before it is "assigned", it will
-    appear in ``local_variables`` and ``global_variables``.
+    Each of the following properties is a dictionary mapping variable, tag or filter
+    names to a list of tuples. Each tuple holds the location of a reference to the name
+    as (<template name>, <line number>). If a name is referenced multiple times, it will
+    appear multiple times in the list. If a name is referenced before it is "assigned",
+    it will appear in ``local_variables`` and ``global_variables``.
 
     :ivar variables: All referenced variables, whether they are in scope or not.
         Including references to names such as ``forloop`` from the ``for`` tag.
@@ -524,6 +555,8 @@ class TemplateAnalysis:
         be visited, probably because they do not implement a ``children`` method.
     :ivar unloadable_partials: Names or identifiers of partial templates that could not
         be loaded. This will be empty if ``follow_partials`` is ``False``.
+    :ivar filters: All filters found during static analysis.
+    :ivar tags: All tags found during static analysis.
     """
 
     __slots__ = (
@@ -532,6 +565,8 @@ class TemplateAnalysis:
         "global_variables",
         "failed_visits",
         "unloadable_partials",
+        "filters",
+        "tags",
     )
 
     def __init__(
@@ -540,14 +575,18 @@ class TemplateAnalysis:
         variables: Refs,
         local_variables: Refs,
         global_variables: Refs,
-        failed_visits: Dict[str, List[Tuple[str, int]]],
-        unloadable_partials: Dict[str, List[Tuple[str, int]]],
+        failed_visits: NameRefs,
+        unloadable_partials: NameRefs,
+        filters: NameRefs,
+        tags: NameRefs,
     ) -> None:
         self.variables = variables
         self.local_variables = local_variables
         self.global_variables = global_variables
         self.failed_visits = failed_visits
         self.unloadable_partials = unloadable_partials
+        self.filters = filters
+        self.tags = tags
 
 
 # pylint: disable=too-many-instance-attributes
@@ -577,30 +616,35 @@ class _TemplateVariableCounter:
         follow_partials: bool = True,
         raise_for_failures: bool = True,
         scope: Optional[ReadOnlyChainMap] = None,
-        template_locals: Optional[ReferenceMap] = None,
+        template_locals: Optional[IdentifierMap] = None,
         partials: Optional[List[Tuple[str, Optional[Dict[str, str]]]]] = None,
     ) -> None:
         self.template = template
         self._template_name = self.template.name or "<string>"
-
         self.follow_partials = follow_partials
         self.raise_for_failures = raise_for_failures
 
         # Names that are added to the template "local" scope.
-        self.template_locals: ReferenceMap = (
+        self.template_locals: IdentifierMap = (
             template_locals if template_locals is not None else defaultdict(list)
         )
+
         # Names that are referenced but are not in the template local scope
-        self.template_globals: ReferenceMap = defaultdict(list)
+        self.template_globals: IdentifierMap = defaultdict(list)
+
         # Names that are referenced by a Liquid expression.
-        self.variables: ReferenceMap = defaultdict(list)
+        self.variables: IdentifierMap = defaultdict(list)
+
+        # Tag and filter names.
+        self.filters: NameRefs = defaultdict(list)
+        self.tags: NameRefs = defaultdict(list)
 
         # Nodes and Expressions that don't implement a `children()` method.
-        self.failed_visits: NameMap = defaultdict(list)
+        self.failed_visits: NameRefs = defaultdict(list)
 
         # Tags that load templates with an expression that can not be analyzed
         # statically.
-        self.unloadable_partials: NameMap = defaultdict(list)
+        self.unloadable_partials: NameRefs = defaultdict(list)
 
         # Block scoped names.
         self._scope = scope if scope is not None else ReadOnlyChainMap()
@@ -631,6 +675,8 @@ class _TemplateVariableCounter:
         return self
 
     def _analyze(self, root: Node) -> None:
+        self._count_tag(root)
+
         try:
             children = root.children()
         except NotImplementedError:
@@ -662,6 +708,8 @@ class _TemplateVariableCounter:
                 self._scope.pop()
 
     async def _analyze_async(self, root: Node) -> None:
+        self._count_tag(root)
+
         try:
             children = root.children()
         except NotImplementedError:
@@ -701,13 +749,13 @@ class _TemplateVariableCounter:
                 self.failed_visits[name].append((self._template_name, child.linenum))
                 return
 
-            for ref in refs:
+            for ref in refs.variable_references:
                 self.variables[ref].append((self._template_name, child.linenum))
 
             # Check refs that are not in scope or in the local namespace before
             # pushing the next block scope. This should highlight names that are
             # expected to be "global".
-            for ref in refs:
+            for ref in refs.variable_references:
                 _ref = RE_SPLIT_IDENT.split(str(ref), 1)[0]
                 if (
                     _ref not in self._scope
@@ -718,18 +766,24 @@ class _TemplateVariableCounter:
                         (self._template_name, child.linenum)
                     )
 
+            for f_ref in refs.filter_references:
+                self.filters[f_ref].append((self._template_name, child.linenum))
+
         if child.template_scope:
             for name in child.template_scope:
                 self.template_locals[
                     Identifier(path=[IdentifierPathElement(name)])
                 ].append((self._template_name, child.linenum))
 
-    def _analyze_expression(self, expression: Expression) -> List[Identifier]:
+    def _analyze_expression(self, expression: Expression) -> References:
         """Return a list of references used in the given expression."""
-        refs: List[Identifier] = []
+        refs: References = References()
 
         if isinstance(expression, Identifier):
-            refs.append(expression)
+            refs.append_variable(expression)
+
+        if isinstance(expression, FilteredExpression):
+            refs.append_filters([f.name for f in expression.filters])
 
         for expr in expression.children():
             refs.extend(self._analyze_expression(expr))
@@ -872,7 +926,7 @@ class _TemplateVariableCounter:
         load_context = child.load_context or {}
 
         # Keep track of partial templates that have already been analyzed. This prevents
-        # us from analysing the same template twice and protects us against recursive
+        # us from analyzing the same template twice and protects us against recursive
         # includes/renders.
         if (name, load_context) in self._partials:
             return None, None
@@ -880,6 +934,11 @@ class _TemplateVariableCounter:
         self._partials.append((name, load_context))
         return name, load_context
 
+    def _count_tag(self, node: Node) -> None:
+        token = node.token()
+        if not isinstance(node, BlockNode) and token.type == TOKEN_TAG:
+            self.tags[token.value].append((self._template_name, token.linenum))
+
     def _update_reference_counters(self, refs: _TemplateVariableCounter) -> None:
         # Accumulate references from the partial/child template into its parent.
         for _name, _refs in refs.variables.items():
@@ -894,6 +953,12 @@ class _TemplateVariableCounter:
         for template_name, _refs in refs.unloadable_partials.items():
             self.unloadable_partials[template_name].extend(_refs)
 
+        for filter_name, _refs in refs.filters.items():
+            self.filters[filter_name].extend(_refs)
+
+        for tag_name, _refs in refs.tags.items():
+            self.tags[tag_name].extend(_refs)
+
     def _raise_for_failures(self) -> None:
         if self.raise_for_failures and self.failed_visits:
             msg_target = next(iter(self.failed_visits.keys()))
@@ -917,3 +982,24 @@ class _TemplateVariableCounter:
             else:
                 msg = f"partial template '{msg_target}' could not be loaded"
             raise TemplateTraversalError(f"failed visit: {msg}")
+
+
+class References:
+    "Collects references for Template.analyze and friends"
+
+    def __init__(self) -> None:
+        self.variable_references: List[Identifier] = []
+        self.filter_references: List[str] = []
+
+    def append_variable(self, var: Identifier) -> None:
+        "Add a variable reference"
+        self.variable_references.append(var)
+
+    def append_filters(self, filters: List[str]) -> None:
+        "Add references to filters"
+        self.filter_references.extend(filters)
+
+    def extend(self, refs: References) -> None:
+        "Incorporate references from another References"
+        self.variable_references.extend(refs.variable_references)
+        self.filter_references.extend(refs.filter_references)
