@@ -1,24 +1,36 @@
-"""Test cases for statically counting a template's variables."""
+"""Test cases for statically counting a template using `BoundTemplate.analyze`."""
 # pylint: disable=too-many-lines
 import asyncio
 
-from typing import List, Optional, TextIO
+from typing import List
+from typing import Optional
+from typing import TextIO
+
 from unittest import TestCase
 
 from liquid import Environment
 from liquid import DictLoader
 from liquid import Template
 
-from liquid.ast import ChildNode, Node
+from liquid.ast import ChildNode
+from liquid.ast import Node
+
 from liquid.context import Context
 from liquid.exceptions import TemplateTraversalError
 from liquid.expression import Expression
+
 from liquid.extra import add_inline_expression_tags
 from liquid.extra import WithTag
+
 from liquid.mode import Mode
 from liquid.stream import TokenStream
 from liquid.tag import Tag
-from liquid.template import BoundTemplate, Refs
+
+from liquid.template import BoundTemplate
+from liquid.template import NameRefs
+from liquid.template import Refs
+from liquid.template import TemplateAnalysis
+
 from liquid.token import Token
 
 
@@ -72,8 +84,8 @@ class MockChildTag(MockTag):
 
 
 # pylint: disable=too-many-public-methods
-class CountTemplateVariablesTestCase(TestCase):
-    """Test that we can count a template's variable references."""
+class AnalyzeTemplateTestCase(TestCase):
+    """Test that we can count a template's variables, tags and filters."""
 
     # pylint: disable=too-many-arguments
     def _test(
@@ -85,26 +97,41 @@ class CountTemplateVariablesTestCase(TestCase):
         failed_visits: Optional[Refs] = None,
         unloadable: Optional[Refs] = None,
         raise_for_failures: bool = True,
+        template_filters: Optional[NameRefs] = None,
+        template_tags: Optional[NameRefs] = None,
     ) -> None:
-        failed_visits = failed_visits if failed_visits is not None else {}
-        unloadable = unloadable if unloadable is not None else {}
-
-        refs = template.analyze(raise_for_failures=raise_for_failures)
-        self.assertEqual(refs.local_variables, template_locals)
-        self.assertEqual(refs.global_variables, template_globals)
-        self.assertEqual(refs.variables, template_refs)
-        self.assertEqual(refs.failed_visits, failed_visits)
-        self.assertEqual(refs.unloadable_partials, unloadable)
+        """Template analysis test helper."""
 
         async def coro():
             return await template.analyze_async(raise_for_failures=raise_for_failures)
 
-        refs = asyncio.run(coro())
-        self.assertEqual(refs.local_variables, template_locals)
-        self.assertEqual(refs.global_variables, template_globals)
-        self.assertEqual(refs.variables, template_refs)
-        self.assertEqual(refs.failed_visits, failed_visits)
-        self.assertEqual(refs.unloadable_partials, unloadable)
+        def assert_refs(refs: TemplateAnalysis) -> None:
+            self.assertEqual(refs.local_variables, template_locals)
+            self.assertEqual(refs.global_variables, template_globals)
+            self.assertEqual(refs.variables, template_refs)
+
+            if failed_visits:
+                self.assertEqual(refs.failed_visits, failed_visits)
+            else:
+                self.assertEqual(refs.failed_visits, {})
+
+            if unloadable:
+                self.assertEqual(refs.unloadable_partials, unloadable)
+            else:
+                self.assertEqual(refs.unloadable_partials, {})
+
+            if template_filters:
+                self.assertEqual(refs.filters, template_filters)
+            else:
+                self.assertEqual(refs.filters, {})
+
+            if template_tags:
+                self.assertEqual(refs.tags, template_tags)
+            else:
+                self.assertEqual(refs.tags, {})
+
+        assert_refs(template.analyze(raise_for_failures=raise_for_failures))
+        assert_refs(asyncio.run(coro()))
 
     def test_analyze_output(self):
         """Test that we can count references in an output statement."""
@@ -122,11 +149,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "z": [("<string>", 1)],
         }
 
+        expected_filters = {
+            "default": [("<string>", 1)],
+        }
+
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_filters,
         )
 
     def test_analyze_identifier_with_bracketed_string_literal(self):
@@ -184,12 +216,16 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"y": [("<string>", 1)], "z": [("<string>", 1)]}
         expected_template_locals = {"x": [("<string>", 1)]}
         expected_refs = {"y": [("<string>", 1)], "z": [("<string>", 1)]}
+        expected_filters = {"append": [("<string>", 1)]}
+        expected_tags = {"assign": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_capture_tag(self):
@@ -199,12 +235,17 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"y": [("<string>", 1)]}
         expected_template_locals = {"x": [("<string>", 1)]}
         expected_refs = {"y": [("<string>", 1)]}
+        expected_tags = {
+            "capture": [("<string>", 1)],
+            "if": [("<string>", 1)],
+        }
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_case_tag(self):
@@ -239,12 +280,14 @@ class CountTemplateVariablesTestCase(TestCase):
             "z": [("<string>", 4)],
             "b": [("<string>", 5)],
         }
+        expected_tags = {"case": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_cycle_tag(self):
@@ -252,22 +295,22 @@ class CountTemplateVariablesTestCase(TestCase):
         template = Template("{% cycle x: a, b %}")
 
         expected_template_globals = {
-            "x": [("<string>", 1)],
             "a": [("<string>", 1)],
             "b": [("<string>", 1)],
         }
         expected_template_locals = {}
         expected_refs = {
-            "x": [("<string>", 1)],
             "a": [("<string>", 1)],
             "b": [("<string>", 1)],
         }
+        expected_tags = {"cycle": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_decrement_tag(self):
@@ -277,12 +320,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {}
         expected_template_locals = {"x": [("<string>", 1)]}
         expected_refs = {}
+        expected_tags = {"decrement": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_echo_tag(self):
@@ -300,12 +345,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "y": [("<string>", 1)],
             "z": [("<string>", 1)],
         }
+        expected_template_filters = {"default": [("<string>", 1)]}
+        expected_tags = {"echo": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_for_tag(self):
@@ -332,11 +381,18 @@ class CountTemplateVariablesTestCase(TestCase):
             "z": [("<string>", 5)],
         }
 
+        expected_tags = {
+            "break": [("<string>", 3)],
+            "continue": [("<string>", 6)],
+            "for": [("<string>", 1)],
+        }
+
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_if_tag(self):
@@ -366,12 +422,14 @@ class CountTemplateVariablesTestCase(TestCase):
             "y": [("<string>", 3)],
             "b": [("<string>", 4)],
         }
+        expected_tags = {"if": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_ifchanged_tag(self):
@@ -381,12 +439,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"x": [("<string>", 1)]}
         expected_template_locals = {}
         expected_refs = {"x": [("<string>", 1)]}
+        expected_tags = {"ifchanged": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_increment_tag(self):
@@ -396,12 +456,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {}
         expected_template_locals = {"x": [("<string>", 1)]}
         expected_refs = {}
+        expected_tags = {"increment": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_liquid_tag(self):
@@ -433,12 +495,23 @@ class CountTemplateVariablesTestCase(TestCase):
             "foo": [("<string>", 3)],
             "i": [("<string>", 9)],
         }
+        expected_template_filters = {
+            "upcase": [("<string>", 3), ("<string>", 5)],
+        }
+        expected_tags = {
+            "echo": [("<string>", 3), ("<string>", 5), ("<string>", 9)],
+            "for": [("<string>", 8)],
+            "if": [("<string>", 2)],
+            "liquid": [("<string>", 1)],
+        }
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_tablerow_tag(self):
@@ -454,12 +527,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "x": [("<string>", 1)],
             "a": [("<string>", 1)],
         }
+        expected_template_filters = {"append": [("<string>", 1)]}
+        expected_tags = {"tablerow": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_unless_tag(self):
@@ -489,12 +566,14 @@ class CountTemplateVariablesTestCase(TestCase):
             "y": [("<string>", 3)],
             "b": [("<string>", 4)],
         }
+        expected_tags = {"unless": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_include_tag(self):
@@ -506,12 +585,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"y": [("some_name", 1)]}
         expected_template_locals = {}
         expected_refs = {"y": [("some_name", 1)]}
+        expected_tags = {"include": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_include_tag_with_assign(self):
@@ -524,12 +605,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"y": [("some_name", 1)]}
         expected_template_locals = {"z": [("some_name", 1)]}
         expected_refs = {"y": [("some_name", 1)], "z": [("<string>", 1)]}
+        expected_tags = {"assign": [("some_name", 1)], "include": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_include_tag_once(self):
@@ -541,12 +624,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"y": [("some_name", 1)]}
         expected_template_locals = {}
         expected_refs = {"y": [("some_name", 1)]}
+        expected_tags = {"include": [("<string>", 1), ("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_recursive_include_tag(self):
@@ -558,12 +643,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"y": [("some_name", 1)]}
         expected_template_locals = {}
         expected_refs = {"y": [("some_name", 1)]}
+        expected_tags = {"include": [("<string>", 1), ("some_name", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_include_tag_with_bound_variable(self):
@@ -587,12 +674,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "x": [("some_name", 1)],
             "some_name": [("some_name", 1)],
         }
+        expected_template_filters = {"append": [("some_name", 1)]}
+        expected_tags = {"include": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_include_tag_with_alias(self):
@@ -608,12 +699,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "foo": [("<string>", 1)],
             "x": [("some_name", 1)],
         }
+        expected_template_filters = {"append": [("some_name", 1)]}
+        expected_tags = {"include": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_include_tag_with_arguments(self):
@@ -631,12 +726,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "x": [("some_name", 1), ("<string>", 2)],
             "y": [("some_name", 1), ("<string>", 1)],
         }
+        expected_template_filters = {"append": [("some_name", 1)]}
+        expected_tags = {"include": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_include_with_variable_name(self):
@@ -651,6 +750,7 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_locals = {}
         expected_refs = {"somevar": [("<string>", 1)], "y": [("<string>", 1)]}
         expected_unloadable_partials = {"somevar": [("<string>", 1)]}
+        expected_tags = {"include": [("<string>", 1)]}
 
         self._test(
             template,
@@ -659,6 +759,7 @@ class CountTemplateVariablesTestCase(TestCase):
             expected_template_globals,
             unloadable=expected_unloadable_partials,
             raise_for_failures=False,
+            template_tags=expected_tags,
         )
 
         with self.assertRaises(TemplateTraversalError):
@@ -680,6 +781,7 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_locals = {}
         expected_refs = {"y": [("<string>", 1)]}
         expected_unloadable_partials = {"nosuchtemplate": [("<string>", 1)]}
+        expected_tags = {"include": [("<string>", 1)]}
 
         self._test(
             template,
@@ -688,6 +790,7 @@ class CountTemplateVariablesTestCase(TestCase):
             expected_template_globals,
             unloadable=expected_unloadable_partials,
             raise_for_failures=False,
+            template_tags=expected_tags,
         )
 
         with self.assertRaises(TemplateTraversalError):
@@ -709,12 +812,17 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"x": [("some_name", 1)], "z": [("some_name", 1)]}
         expected_template_locals = {"z": [("<string>", 1)]}
         expected_refs = {"x": [("some_name", 1)], "z": [("some_name", 1)]}
+        expected_tags = {
+            "assign": [("<string>", 1), ("some_name", 1)],
+            "render": [("<string>", 1)],
+        }
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_render_tag_once(self):
@@ -726,12 +834,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"x": [("some_name", 1)]}
         expected_template_locals = {}
         expected_refs = {"x": [("some_name", 1)]}
+        expected_tags = {"render": [("<string>", 1), ("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_recursive_render_tag(self):
@@ -743,12 +853,14 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_globals = {"y": [("some_name", 1)]}
         expected_template_locals = {}
         expected_refs = {"y": [("some_name", 1)]}
+        expected_tags = {"render": [("<string>", 1), ("some_name", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_analyze_render_tag_with_bound_variable(self):
@@ -770,12 +882,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "x": [("some_name", 1)],
             "some_name": [("some_name", 1)],
         }
+        expected_template_filters = {"append": [("some_name", 1)]}
+        expected_tags = {"render": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_render_tag_with_alias(self):
@@ -791,12 +907,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "foo": [("<string>", 1)],
             "x": [("some_name", 1)],
         }
+        expected_template_filters = {"append": [("some_name", 1)]}
+        expected_tags = {"render": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_render_tag_with_arguments(self):
@@ -814,12 +934,16 @@ class CountTemplateVariablesTestCase(TestCase):
             "x": [("some_name", 1), ("<string>", 2)],
             "y": [("some_name", 1), ("<string>", 1)],
         }
+        expected_template_filters = {"append": [("some_name", 1)]}
+        expected_tags = {"render": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
+            template_tags=expected_tags,
         )
 
     def test_analyze_render_tag_scope(self):
@@ -839,12 +963,17 @@ class CountTemplateVariablesTestCase(TestCase):
             "z": [("some_name", 1)],
             "y": [("<string>", 1)],
         }
+        expected_tags = {
+            "assign": [("<string>", 1), ("some_name", 1)],
+            "render": [("<string>", 1)],
+        }
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_render_template_not_found(self):
@@ -856,6 +985,7 @@ class CountTemplateVariablesTestCase(TestCase):
         expected_template_locals = {}
         expected_refs = {"y": [("<string>", 1)]}
         expected_unloadable_partials = {"nosuchtemplate": [("<string>", 1)]}
+        expected_tags = {"render": [("<string>", 1)]}
 
         self._test(
             template,
@@ -864,6 +994,7 @@ class CountTemplateVariablesTestCase(TestCase):
             expected_template_globals,
             unloadable=expected_unloadable_partials,
             raise_for_failures=False,
+            template_tags=expected_tags,
         )
 
         with self.assertRaises(TemplateTraversalError):
@@ -883,6 +1014,7 @@ class CountTemplateVariablesTestCase(TestCase):
         template = env.from_string("{% mock %}\n{% mock %}")
 
         expected_failed_visits = {"MockNode": [("<string>", 1), ("<string>", 2)]}
+        expected_tags = {"mock": [("<string>", 1), ("<string>", 2)]}
 
         self._test(
             template,
@@ -891,6 +1023,7 @@ class CountTemplateVariablesTestCase(TestCase):
             {},
             failed_visits=expected_failed_visits,
             raise_for_failures=False,
+            template_tags=expected_tags,
         )
 
         with self.assertRaises(TemplateTraversalError):
@@ -910,6 +1043,7 @@ class CountTemplateVariablesTestCase(TestCase):
         template = env.from_string("{% mock %}\n{% mock %}")
 
         expected_failed_visits = {"MockExpression": [("<string>", 1), ("<string>", 2)]}
+        expected_tags = {"mock": [("<string>", 1), ("<string>", 2)]}
 
         self._test(
             template,
@@ -918,6 +1052,7 @@ class CountTemplateVariablesTestCase(TestCase):
             {},
             failed_visits=expected_failed_visits,
             raise_for_failures=False,
+            template_tags=expected_tags,
         )
 
         with self.assertRaises(TemplateTraversalError):
@@ -950,12 +1085,17 @@ class CountTemplateVariablesTestCase(TestCase):
             "bar": [("<string>", 1)],
             "baz": [("<string>", 1)],
         }
+        expected_template_filters = {
+            "upcase": [("<string>", 1)],
+            "append": [("<string>", 1)],
+        }
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_filters=expected_template_filters,
         )
 
     def test_analyze_with_tag(self) -> None:
@@ -982,12 +1122,14 @@ class CountTemplateVariablesTestCase(TestCase):
             "collection.products.first.title": [("<string>", 1)],
             "p.title": [("<string>", 1), ("<string>", 1)],
         }
+        expected_tags = {"with": [("<string>", 1)]}
 
         self._test(
             template,
             expected_refs,
             expected_template_locals,
             expected_template_globals,
+            template_tags=expected_tags,
         )
 
     def test_variable_parts(self) -> None:
